mod fuse_daemon;
use fuse_daemon::FuseDaemon;
mod fuse_fs;

use btfproto::{local_fs::LocalFs, server::FsProvider};
use btlib::{
    crypto::{
        tpm::{TpmCredStore, TpmCreds},
        CredStore, Creds, CredsPriv,
    },
    BlockPath, Result,
};
use core::future::Future;
use log::{error};
use serde_json::from_str;
use std::{
    collections::HashMap,
    env::VarError,
    ffi::OsString,
    fs::{self},
    io,
    path::{Path, PathBuf},
    str::FromStr,
    sync::Arc,
};
use tss_esapi::{
    tcti_ldr::{TabrmdConfig, TctiNameConf},
    Context,
};

const DEFAULT_TABRMD: &str = "bus_type=session";
const MOUNT_OPTIONS: &str = "default_permissions";
const TABRMD_ENVVAR: &str = "BT_TABRMD";
const UIDMAP_ENVVAR: &str = "BT_UIDMAP";

trait PathExt {
    fn try_create_dir(&self) -> io::Result<()>;
}

impl<T: AsRef<Path>> PathExt for T {
    fn try_create_dir(&self) -> io::Result<()> {
        match fs::create_dir(self) {
            Ok(_) => Ok(()),
            Err(err) => match err.kind() {
                io::ErrorKind::AlreadyExists => Ok(()),
                _ => Err(err),
            },
        }
    }
}

fn tpm_state_path<T: AsRef<Path>>(main_dir: T) -> PathBuf {
    main_dir.as_ref().join("tpm_state")
}

fn block_dir<T: AsRef<Path>>(main_dir: T) -> PathBuf {
    main_dir.as_ref().join("bt")
}

fn parse_uid_map(uid_map: &str) -> Result<HashMap<u32, Arc<BlockPath>>> {
    let map: HashMap<u32, BlockPath> = from_str(uid_map)?;
    let mut output = HashMap::with_capacity(map.len());
    for (key, value) in map.into_iter() {
        output.insert(key, Arc::new(value));
    }
    Ok(output)
}

fn node_creds(state_file: PathBuf, tabrmd_cfg: &str) -> Result<TpmCreds> {
    let context = Context::new(TctiNameConf::Tabrmd(TabrmdConfig::from_str(tabrmd_cfg)?))?;
    let cred_store = TpmCredStore::new(context, state_file)?;
    cred_store.node_creds()
}

fn provider<C: 'static + Creds + Send + Sync>(
    btdir: PathBuf,
    node_creds: C,
) -> Result<impl FsProvider> {
    btdir.try_create_dir()?;
    let empty = fs::read_dir(&btdir)?.next().is_none();
    if empty {
        LocalFs::new_empty(btdir, 0, node_creds, btfproto::local_fs::ModeAuthorizer {})
    } else {
        LocalFs::new_existing(btdir, node_creds, btfproto::local_fs::ModeAuthorizer {})
    }
}

fn run_daemon(
    main_dir: OsString,
    tabrmd_string: Option<String>,
    uid_map: Option<String>,
    tpm_state_file: Option<PathBuf>,
) -> impl Send + Sync + Future<Output = ()> {
    let main_dir = PathBuf::from(main_dir);
    let tabrmd_cfg = tabrmd_string
        .as_ref()
        .map_or(DEFAULT_TABRMD, |s| s.as_str());
    let uid_map = {
        let uid_map = uid_map.unwrap_or_else(|| "{}".to_owned());
        parse_uid_map(&uid_map).unwrap()
    };
    let tpm_state_file = tpm_state_file.unwrap_or_else(|| tpm_state_path(&main_dir));
    let node_creds = node_creds(tpm_state_file, tabrmd_cfg).expect("failed to get node creds");
    let fallback_path = {
        let writecap = node_creds
            .writecap()
            .ok_or(btlib::BlockError::MissingWritecap)
            .unwrap();
        Arc::new(writecap.bind_path())
    };
    let provider = {
        let btdir = block_dir(&main_dir);
        provider(btdir, node_creds).expect("failed to create FS provider")
    };

    let mut daemon = FuseDaemon::new(main_dir, uid_map, fallback_path, provider)
        .expect("failed to create FUSE daemon");
    async move { daemon.finished().await }
}

#[tokio::main]
async fn main() {
    env_logger::init();
    let main_dir = std::env::args_os().nth(1).expect("no mount point given");
    let tabrmd_string = std::env::var(TABRMD_ENVVAR).ok();
    let uid_map = match std::env::var(UIDMAP_ENVVAR) {
        Ok(uid_map) => Some(uid_map),
        Err(err) => match err {
            VarError::NotPresent => None,
            VarError::NotUnicode(_) => {
                error!("environment variable {UIDMAP_ENVVAR} contained non-unicode characters");
                return;
            }
        },
    };
    run_daemon(main_dir, tabrmd_string, uid_map, None).await;
}

#[cfg(test)]
mod test {
    use super::*;

    use btlib::{crypto::Creds, log::BuilderExt, Epoch, Principaled};
    use ctor::ctor;
    use std::{
        ffi::{OsStr, OsString},
        fs::{
            create_dir, hard_link, metadata, read, read_dir, remove_dir, remove_file, rename,
            set_permissions, write, Permissions, ReadDir,
        },
        os::unix::fs::PermissionsExt,
        thread::JoinHandle,
        time::Duration,
        sync::mpsc,
    };
    use swtpm_harness::SwtpmHarness;
    use tempdir::TempDir;

    /// An optional timeout to wait for the FUSE daemon to start in tests.
    const TIMEOUT: Option<Duration> = Some(Duration::from_millis(1000));

    /// The log level to use when running tests.
    /// Note that the debug log level significantly reduces performance.
    const LOG_LEVEL: &str = "warn";

    #[ctor]
    fn ctor() {
        std::env::set_var("RUST_LOG", LOG_LEVEL);
        env_logger::Builder::from_default_env().btformat().init();
    }

    /// Unmounts the file system at the given path.
    fn unmount<P: AsRef<Path>>(mnt_path: P) {
        const PROG: &str = "fusermount";
        let mnt_path = mnt_path
            .as_ref()
            .as_os_str()
            .to_str()
            .expect("failed to convert mnt_path to `str`");
        let code = std::process::Command::new(PROG)
            .args(["-u", mnt_path])
            .status()
            .expect("waiting for exit status failed")
            .code()
            .expect("code returned None");
        if code != 0 {
            panic!("{PROG} exited with a non-zero status: {code}");
        }
    }

    fn file_names(read_dir: ReadDir) -> impl Iterator<Item = OsString> {
        read_dir.map(|entry| entry.unwrap().file_name())
    }

    const ROOT_PASSWD: &str = "password";

    struct TestCase {
        _temp_dir: TempDir,
        mnt_path: PathBuf,
        handle: Option<JoinHandle<()>>,
    }

    impl TestCase {
        fn new() -> TestCase {
            let tmp = TempDir::new("btfuse").unwrap();
            let mnt_path = FuseDaemon::mnt_path(tmp.path());
            let main_dir = OsString::from(tmp.path().to_owned());
            let (mounted_tx, mounted_rx) = mpsc::channel();
            let handle = std::thread::spawn(move || Self::run(main_dir, mounted_tx));
            match TIMEOUT {
                Some(duration) => mounted_rx.recv_timeout(duration).unwrap(),
                None => mounted_rx.recv().unwrap(),
            };
            Self {
                _temp_dir: tmp,
                mnt_path,
                handle: Some(handle),
            }
        }

        fn run(main_dir: OsString, mounted_tx: mpsc::Sender<()>) {
            let swtpm = Self::swtpm();
            let state_path = swtpm.state_path().to_owned();
            let tabrmd_string = Some(swtpm.tabrmd_config().to_owned());
            let uid_map = None;
            let runtime = tokio::runtime::Builder::new_current_thread()
                .build()
                .unwrap();
            // run_daemon can only be called in the context of a runtime, hence the need to call
            // spawn_blocking.
            let started = runtime.spawn_blocking(|| run_daemon(
                main_dir,
                tabrmd_string,
                uid_map,
                Some(state_path),
            ));
            let future = runtime.block_on(started).unwrap();

            // The file system is mounted before run_daemon returns.
            mounted_tx.send(()).unwrap();
            runtime.block_on(future);
        }

        fn swtpm() -> SwtpmHarness {
            let swtpm = SwtpmHarness::new().unwrap();
            let state_path: PathBuf = swtpm.state_path().to_owned();
            let cred_store = {
                let context = swtpm.context().unwrap();
                TpmCredStore::new(context, state_path.clone()).unwrap()
            };
            let root_creds = cred_store.gen_root_creds(ROOT_PASSWD).unwrap();
            let mut node_creds = cred_store.node_creds().unwrap();
            let expires = Epoch::now() + Duration::from_secs(3600);
            let writecap = root_creds
                .issue_writecap(node_creds.principal(), vec![], expires)
                .unwrap();
            cred_store
                .assign_node_writecap(&mut node_creds, writecap)
                .unwrap();
            swtpm
        }

        fn mnt_path(&self) -> &PathBuf {
            &self.mnt_path
        }

        fn wait(&mut self) {
            if let Some(handle) = self.handle.take() {
                handle.join().expect("join failed");
            }
        }

        fn unmount_and_wait(&mut self) {
            // If `handle` has already been taken that means `wait` has already been called. If
            // this thread called `wait` and subsequently became unblocked, we know that the FUSE
            // thread halted, which only happens when the file system is unmounted. Hence
            // we don't have to unmount it, nor wait for the FUSE thread.
            if self.handle.is_none() {
                return;
            }
            let mnt_path = self.mnt_path();
            unmount(mnt_path);
            self.wait();
        }
    }

    impl Drop for TestCase {
        fn drop(&mut self) {
            self.unmount_and_wait()
        }
    }

    /// Creates a new file system and mounts it at `/tmp/btfuse.<random>/mnt` so it can be
    /// tested manually.
    //#[test]
    #[allow(dead_code)]
    fn manual_test() {
        let mut case = TestCase::new();
        case.wait()
    }

    /// Tests if the file system can be mount then unmounted successfully.
    #[test]
    fn mount_then_unmount() {
        let _ = TestCase::new();
    }

    #[test]
    fn write_read() -> Result<()> {
        const EXPECTED: &[u8] =
            b"The paths to failure are uncountable, yet to success there is but one.";
        let case = TestCase::new();
        let file_path = case.mnt_path().join("file");

        write(&file_path, EXPECTED)?;

        let actual = read(&file_path)?;
        assert_eq!(EXPECTED, actual);
        Ok(())
    }

    #[test]
    fn create_file_then_readdir() {
        const DATA: &[u8] = b"Au revoir Shoshanna!";
        let file_name = OsStr::new("landa_dialog.txt");
        let expected = [file_name];
        let case = TestCase::new();
        let mnt_path = case.mnt_path();
        let file_path = mnt_path.join(file_name);

        write(&file_path, DATA).expect("write failed");

        let first = file_names(read_dir(&mnt_path).expect("read_dir failed"));
        assert!(first.eq(expected));
        let second = file_names(read_dir(&mnt_path).expect("read_dir failed"));
        assert!(second.eq(expected));
    }

    #[test]
    fn create_then_delete_file() {
        const DATA: &[u8] = b"The universe is hostile, so impersonal. Devour to survive";
        let file_name = OsStr::new("tool_lyrics.txt");
        let case = TestCase::new();
        let mnt_path = case.mnt_path();
        let file_path = mnt_path.join(file_name);
        write(&file_path, DATA).expect("write failed");

        remove_file(&file_path).expect("remove_file failed");

        let expected: [&OsStr; 0] = [];
        assert!(file_names(read_dir(&mnt_path).expect("read_dir failed")).eq(expected))
    }

    //#[test]
    //fn hard_link_then_remove() {
    //    const EXPECTED: &[u8] = b"And the lives we've reclaimed";
    //    let name1 = OsStr::new("refugee_lyrics.txt");
    //    let name2 = OsStr::new("rise_against_lyrics.txt");
    //    let case = TestCase::new();
    //    let mnt_path = case.mnt_path();
    //    let path1 = mnt_path.join(name1);
    //    let path2 = mnt_path.join(name2);
    //    write(&path1, EXPECTED).expect("write failed");

    //    hard_link(&path1, &path2).expect("hard_link failed");
    //    remove_file(&path1).expect("remove_file failed");

    //    let actual = read(&path2).expect("read failed");
    //    assert_eq!(EXPECTED, actual);
    //}

    //#[test]
    //fn hard_link_then_remove_both() {
    //    const EXPECTED: &[u8] = b"And the lives we've reclaimed";
    //    let name1 = OsStr::new("refugee_lyrics.txt");
    //    let name2 = OsStr::new("rise_against_lyrics.txt");
    //    let case = TestCase::new();
    //    let mnt_path = case.mnt_path();
    //    let path1 = mnt_path.join(name1);
    //    let path2 = mnt_path.join(name2);
    //    write(&path1, EXPECTED).expect("write failed");

    //    hard_link(&path1, &path2).expect("hard_link failed");
    //    remove_file(&path1).expect("remove_file on path1 failed");
    //    remove_file(&path2).expect("remove_file on path2 failed");

    //    let expected: [&OsStr; 0] = [];
    //    assert!(file_names(read_dir(&mnt_path).expect("read_dir failed")).eq(expected));
    //}

    //#[test]
    //fn set_mode_bits() {
    //    const EXPECTED: u32 = libc::S_IFREG | 0o777;
    //    let case = TestCase::new();
    //    let file_path = case.mnt_path().join("bagobits");
    //    write(&file_path, []).expect("write failed");
    //    let original = metadata(&file_path)
    //        .expect("metadata failed")
    //        .permissions()
    //        .mode();
    //    assert_ne!(EXPECTED, original);

    //    set_permissions(&file_path, Permissions::from_mode(EXPECTED))
    //        .expect("set_permissions failed");

    //    let actual = metadata(&file_path)
    //        .expect("metadata failed")
    //        .permissions()
    //        .mode();
    //    assert_eq!(EXPECTED, actual);
    //}

    //#[test]
    //fn create_directory() {
    //    const EXPECTED: &str = "etc";
    //    let case = TestCase::new();
    //    let mnt_path = case.mnt_path();
    //    let dir_path = mnt_path.join(EXPECTED);

    //    create_dir(&dir_path).expect("create_dir failed");

    //    let actual = file_names(read_dir(mnt_path).expect("read_dir failed"));
    //    assert!(actual.eq([EXPECTED]));
    //}

    //#[test]
    //fn create_file_under_new_directory() {
    //    const DIR_NAME: &str = "etc";
    //    const FILE_NAME: &str = "file";
    //    let case = TestCase::new();
    //    let mnt_path = case.mnt_path();
    //    let dir_path = mnt_path.join(DIR_NAME);
    //    let file_path = dir_path.join(FILE_NAME);

    //    create_dir(&dir_path).expect("create_dir failed");
    //    write(&file_path, []).expect("write failed");

    //    let actual = file_names(read_dir(dir_path).expect("read_dir failed"));
    //    assert!(actual.eq([FILE_NAME]));
    //}

    //#[test]
    //fn create_then_remove_directory() {
    //    const DIR_NAME: &str = "etc";
    //    let case = TestCase::new();
    //    let mnt_path = case.mnt_path();
    //    let dir_path = mnt_path.join(DIR_NAME);

    //    create_dir(&dir_path).expect("create_dir failed");
    //    remove_dir(&dir_path).expect("remove_dir failed");

    //    let actual = file_names(read_dir(&mnt_path).expect("read_dir failed"));
    //    const EMPTY: [&str; 0] = [""; 0];
    //    assert!(actual.eq(EMPTY));
    //}

    //#[test]
    //fn read_only_dir_cant_create_subdir() {
    //    const DIR_NAME: &str = "etc";
    //    let case = TestCase::new();
    //    let dir_path = case.mnt_path().join(DIR_NAME);
    //    create_dir(&dir_path).expect("create_dir failed");
    //    set_permissions(&dir_path, Permissions::from_mode(libc::S_IFDIR | 0o444))
    //        .expect("set_permissions failed");

    //    let result = create_dir(dir_path.join("sub"));

    //    let err = result.err().expect("create_dir returned `Ok`");
    //    let os_err = err.raw_os_error().expect("raw_os_error was empty");
    //    assert_eq!(os_err, libc::EACCES);
    //}

    //#[test]
    //fn read_only_dir_cant_remove_subdir() {
    //    const DIR_NAME: &str = "etc";
    //    let case = TestCase::new();
    //    let dir_path = case.mnt_path().join(DIR_NAME);
    //    let sub_path = dir_path.join("sub");
    //    create_dir(&dir_path).expect("create_dir failed");
    //    create_dir(&sub_path).expect("create_dir failed");
    //    set_permissions(&dir_path, Permissions::from_mode(libc::S_IFDIR | 0o444))
    //        .expect("set_permissions failed");

    //    let result = remove_dir(&sub_path);

    //    let err = result.err().expect("remove_dir returned `Ok`");
    //    let os_err = err.raw_os_error().expect("raw_os_error was empty");
    //    assert_eq!(os_err, libc::EACCES);
    //}

    //#[test]
    //fn rename_file() {
    //    const FILE_NAME: &str = "parabola.txt";
    //    const EXPECTED: &[u8] = b"We are eternal all this pain is an illusion";
    //    let case = TestCase::new();
    //    let src_path = case.mnt_path().join(FILE_NAME);
    //    let dst_path = case.mnt_path().join("parabola_lyrics.txt");

    //    write(&src_path, EXPECTED).unwrap();
    //    rename(&src_path, &dst_path).unwrap();

    //    let actual = read(&dst_path).unwrap();
    //    assert_eq!(EXPECTED, actual)
    //}
}
